@@ -233,6 +233,10 @@ srl_build_decoder_struct(pTHX_ HV *opt, sv_with_hash *options)
         if ( val && SvTRUE(val) )
             SRL_DEC_SET_OPTION(dec, SRL_F_DECODER_REFUSE_ZLIB);
 
+        my_hv_fetchs(he,val,opt, SRL_DEC_OPT_IDX_REFUSE_ZSTD);
+        if ( val && SvTRUE(val) )
+            SRL_DEC_SET_OPTION(dec, SRL_F_DECODER_REFUSE_ZSTD);
+
         my_hv_fetchs(he,val,opt, SRL_DEC_OPT_IDX_REFUSE_OBJECTS);
         if ( val && SvTRUE(val) )
             SRL_DEC_SET_OPTION(dec, SRL_F_DECODER_REFUSE_OBJECTS);
@@ -403,6 +407,9 @@ srl_decode_into_internal(pTHX_ srl_decoder_t *origdec, SV *src, SV *header_into,
     } else if (expect_false( SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_DECOMPRESS_ZLIB) )) {
         dec->bytes_consumed = srl_decompress_body_zlib(aTHX_ dec->pbuf, NULL);
         origdec->bytes_consumed = dec->bytes_consumed;
+    } else if (expect_false( SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_DECOMPRESS_ZSTD) )) {
+        dec->bytes_consumed = srl_decompress_body_zstd(aTHX_ dec->pbuf, NULL);
+        origdec->bytes_consumed = dec->bytes_consumed;
     }
 
     /* this function *MUST* be called right after srl_decompress* functions */
@@ -580,7 +587,7 @@ srl_read_header(pTHX_ srl_decoder_t *dec, SV *header_user_data)
 
         if (expect_false( dec->proto_version == 1 ))
             SRL_DEC_SET_OPTION(dec, SRL_F_DECODER_PROTOCOL_V1); /* compat mode */
-        else if (expect_false( dec->proto_version > 3 || dec->proto_version < 1 ))
+        else if (expect_false( dec->proto_version > SRL_PROTOCOL_VERSION || dec->proto_version < 1 ))
             SRL_RDR_ERRORf1(dec->pbuf, "Unsupported Sereal protocol version %u", dec->proto_version);
 
         if (dec->encoding_flags == SRL_PROTOCOL_ENCODING_RAW) {
@@ -606,6 +613,15 @@ srl_read_header(pTHX_ srl_decoder_t *dec, SV *header_user_data)
             dec->flags |= SRL_F_DECODER_DECOMPRESS_ZLIB;
         }
         else
+        if (dec->encoding_flags == SRL_PROTOCOL_ENCODING_ZSTD)
+        {
+            if (expect_false( SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_REFUSE_ZSTD) )) {
+                SRL_RDR_ERROR(dec->pbuf, "Sereal document is compressed with ZSTD, "
+                              "but this decoder is configured to refuse ZSTD-compressed input.");
+            }
+            dec->flags |= SRL_F_DECODER_DECOMPRESS_ZSTD;
+        }
+        else
         {
             SRL_RDR_ERRORf1(dec->pbuf, "Sereal document encoded in an unknown format '%d'",
                             dec->encoding_flags >> SRL_PROTOCOL_VERSION_BITS);
@@ -1083,7 +1099,7 @@ srl_read_hash(pTHX_ srl_decoder_t *dec, SV* into, U8 tag) {
         else
         if ( expect_false( SvTYPE(*fetched_sv) != SVt_NULL ) ) {
             /* sv_dump(*fetched_sv); */
-            SRL_RDR_ERRORf2(dec->pbuf, "duplicate key '%.*s' in hash", key_len, (char *)from);
+            SRL_RDR_ERRORf2(dec->pbuf, "duplicate key '%.*s' in hash", (int) key_len, (char *)from);
         }
         srl_read_single_value(aTHX_ dec, *fetched_sv, fetched_sv );
     }
