@@ -15,11 +15,38 @@ use Scalar::Util qw(reftype blessed refaddr);
 use Config;
 use Carp qw(confess);
 use Storable qw(dclone);
+use Cwd;
 
 # Dynamically load constants from whatever is being tested
-our ($Class, $ConstClass);
+our ($Class, $ConstClass, $InRepo);
+sub get_git_top_dir {
+    my @dirs = (0, 1, 2, 4);
+    for my $d (@dirs) {
+        my $tdir = File::Spec->catdir(map File::Spec->updir, 1..$d);
+        my $gdir = File::Spec->catdir($tdir, '.git');
+        return $tdir
+            if -d $gdir;
+    }
+    return();
+}
+
+BEGIN{
+    if (defined(my $top_dir = get_git_top_dir())) {
+        for my $need ('Encoder', 'Decoder') {
+            my $blib_dir = File::Spec->catdir($top_dir, 'Perl', $need, "blib");
+            if (-d $blib_dir) {
+                require blib;
+                blib->import($blib_dir);
+            }
+        }
+        $InRepo=1;
+    }
+}
 BEGIN {
-    if (-e "lib/Sereal/Encoder.pm") {
+    if (-e "lib/Sereal.pm") {
+        $Class = 'Sereal::Encoder';
+    }
+    elsif (-e "lib/Sereal/Encoder.pm") {
         $Class = 'Sereal::Encoder';
     }
     elsif (-e "lib/Sereal/Decoder.pm") {
@@ -31,7 +58,7 @@ BEGIN {
     elsif (-e "lib/Sereal/Splitter.pm") {
         $Class = 'Sereal::Splitter';
     } else {
-        die "Could not find an applicable Sereal constants location";
+        die "Could not find an applicable Sereal constants location (in: ",cwd(),")";
     }
     $ConstClass = $Class . "::Constants";
     eval "use $ConstClass ':all'; 1"
@@ -601,16 +628,6 @@ sub setup_tests {
 }
 
 
-sub get_git_top_dir {
-    my @dirs = (0, 1, 2);
-    for my $d (@dirs) {
-        my $tdir = File::Spec->catdir(map File::Spec->updir, 1..$d);
-        my $gdir = File::Spec->catdir($tdir, '.git');
-        return $tdir
-            if -d $gdir;
-    }
-    return();
-}
 
 sub have_encoder_and_decoder {
     my ($min_v)= @_;
@@ -618,13 +635,6 @@ sub have_encoder_and_decoder {
     my $need = $Class =~ /Encoder/ ? "Decoder" : "Encoder";
     my $need_class = "Sereal::$need";
 
-    if (defined(my $top_dir = get_git_top_dir())) {
-        my $blib_dir = File::Spec->catdir($top_dir, 'Perl', $need, "blib");
-        if (-d $blib_dir) {
-            require blib;
-            blib->import($blib_dir);
-        }
-    }
     eval "use $Class; 1"
     or do {
         note("Could not locate $Class for testing" . ($@ ? " (Exception: $@)" : ""));
@@ -638,7 +648,7 @@ sub have_encoder_and_decoder {
     };
     my $cmp_v = $need_class->VERSION;
     if ($min_v and $cmp_v < $min_v) {
-        note("Could not load correct version of $need_class for testing "
+        diag("Could not load correct version of $need_class for testing "
              ."(got: $cmp_v, needed at least $min_v)");
         return;
     }
@@ -646,7 +656,7 @@ sub have_encoder_and_decoder {
     $cmp_v = sprintf("%.2f", int($cmp_v*100)/100);
     my %compat_versions = map {$_ => 1} $Class->_test_compat();
     if (not defined $cmp_v or not exists $compat_versions{$cmp_v}) {
-        note("Could not load correct version of $need_class for testing "
+        diag("Could not load correct version of $need_class for testing "
              ."(got: $cmp_v, needed any of ".join(", ", keys %compat_versions).")");
         return();
     }
@@ -681,6 +691,10 @@ my $eng2= "1e3";
 
 my $sum= $eng0e0 + $eng0e1 + $eng2;
 
+sub encoder_required {
+    my ($ver, $name)= @_;
+    return "" . ( $Sereal::Encoder::VERSION < $ver ? "TODO " : "") . $name;
+}
 
 sub _get_roundtrip_tests {
     my @ScalarRoundtripTests = (
@@ -734,8 +748,8 @@ sub _get_roundtrip_tests {
             1023,1024,1025,
             8191,8192,8193,
         )),
-        ( map { [ ( $Sereal::Encoder::VERSION < 3.005002 ? "TODO" : "")
-                    . " troublesome num/strs '$_'", $_ ] } @numstr ),
+        ( map { [ encoder_required(3.005002, " troublesome num/strs '$_'"),
+                  $_ ] } @numstr ),
         ["long latin1 string", "üll" x 10000],
         ["long utf8 string", do {use utf8; " עדיין חשב" x 10000}],
         ["long utf8 string with only ascii", do {use utf8; "foo" x 10000}],
@@ -810,8 +824,27 @@ sub _get_roundtrip_tests {
         [ "undef", [\undef, \undef] ],
     );
 
+    my @blessed_array_check1;
+    $blessed_array_check1[0]= "foo";
+    $blessed_array_check1[1]= bless \$blessed_array_check1[0], "BlessedArrayCheck";
+    $blessed_array_check1[2]= \$blessed_array_check1[0];
+
+    my @blessed_array_check2= (3,0,0,3);
+    $blessed_array_check2[1]= \$blessed_array_check2[0];
+    $blessed_array_check2[2]= \$blessed_array_check2[3];
+    bless \$blessed_array_check2[0], "BlessedArrayCheck";
+    bless \$blessed_array_check2[3], "BlessedArrayCheck";
+
+    my @sc_array=(1,1);
+    $sc_array[0]=bless \$sc_array[1], "BlessedArrayLeft";
+    $sc_array[1]=bless \$sc_array[0], "BlessedArrayRight";
+
+
     my @RoundtripTests = (
         @ScalarRoundtripTests,
+        [ encoder_required(3.006006,"BlessedArrayCheck 1"), \@blessed_array_check1 ],
+        [ encoder_required(3.006006,"BlessedArrayCheck 2"), \@blessed_array_check2 ],
+        [ encoder_required(3.006006,"Scalar Cross Blessed Array"), \@sc_array ],
 
         ["[{foo => 1}, {foo => 2}] - repeated hash keys",
           [{foo => 1}, {foo => 2}] ],
@@ -1099,7 +1132,8 @@ sub run_roundtrip_tests_internal {
             eval {$decoded2 = $dec->($encoded2); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while encoding the second time: $err");
+                    diag("Got error while decoding the second time: $err");
+                    # hobodecode($encoded2);
                 };
 
             defined($decoded2) == defined($data)
@@ -1127,7 +1161,7 @@ sub run_roundtrip_tests_internal {
             eval {$decoded3 = $dec->($encoded3); 1}
                 or do {
                     my $err = $@ || 'Zombie error';
-                    diag("Got error while encoding the third time: $err");
+                    diag("Got error while decoding the third time: $err");
                 };
 
             defined($decoded3) == defined($data)
@@ -1160,8 +1194,14 @@ sub run_roundtrip_tests_internal {
                         my @clean= ($ename, $name);
                         s/[^\w.-]+/_/g, s/__+/_/g for @clean;
                         my $cleaned= join "/", @clean;
-                        my $dir= $0;
-                        $dir=~s!/[^/]+\z!/data/$clean[0]!;
+                        my ($v,$p,$d)= File::Spec->splitpath($0);
+                        my $dir= File::Spec->catpath(
+                            $v, 
+                            File::Spec->catdir(
+                                File::Spec->splitdir($p),
+                                "data",$clean[0]
+                            )
+                        );
                         mkpath $dir unless -d $dir;
                         my $base= "$dir/$clean[1].enc";
                         $seen_name{$combined_name}= $base;
@@ -1197,13 +1237,13 @@ sub _write_file {
 # For bootstrapping other language implementations' tests
 our $COMPRESS;
 sub write_test_files {
-    my $dir = shift;
+    my ($dir, $version) = @_;
     require File::Path;
     File::Path::mkpath($dir);
     my $make_data_file_name = sub {File::Spec->catfile($dir, sprintf("test_data_%05u", shift))};
     my $make_name_file_name = sub {File::Spec->catfile($dir, sprintf("test_name_%05u", shift))};
 
-    setup_tests();
+    setup_tests($version);
     foreach my $testno (1..@BasicTests) {
         my $t = $BasicTests[$testno-1];
         my $data = ref($t->[1]) eq 'CODE' ? $t->[1]->() : $t->[1];
